
//	HID Minidriver for Microsoft Xbox 360 Wireless Controller

#include	"driver.h"

//----------------------------------------------------------------------------------------------
//	萔
//----------------------------------------------------------------------------------------------

#define	INTERVAL					-1
#define	INTERVAL_WAIT				-4 * 1000 * 10
#define	INTERVAL_STALL				-16 * 1000 * 10

//----------------------------------------------------------------------------------------------
//	֐
//----------------------------------------------------------------------------------------------

NTSTATUS StartDevice( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
NTSTATUS QueryCapabilities( IN PDEVICE_OBJECT, IN PIRP, IN PIO_STACK_LOCATION );
NTSTATUS SurpriseRemoval( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
NTSTATUS RemoveDevice( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
KSTART_ROUTINE WorkerThreadIn;
KSTART_ROUTINE WorkerThreadOut;
KSTART_ROUTINE WorkerThreadIdc;

//----------------------------------------------------------------------------------------------
//	PnP
//----------------------------------------------------------------------------------------------
NTSTATUS PnP(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ϐ錾
	NTSTATUS			Status;
	PDEVICE_EXTENSION	DeviceExtension	= GET_MINIDRIVER_DEVICE_EXTENSION( DeviceObject );
	PIO_STACK_LOCATION	CurrentIrpStack;

	//	QƃJE^Z
	Status	= IncrementReference( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		CompleteRequest( Irp, Status, 0 );
		return( Status );
	}

	//	}Ci[ t@NVɂďU蕪
	CurrentIrpStack	= IoGetCurrentIrpStackLocation( Irp );
	switch( CurrentIrpStack->MinorFunction )
	{
		case IRP_MN_START_DEVICE:
			Status	= StartDevice( DeviceObject, DeviceExtension, Irp );
			break;

		case IRP_MN_QUERY_CAPABILITIES:
			Status	= QueryCapabilities( DeviceObject, Irp, CurrentIrpStack );
			break;

		case IRP_MN_SURPRISE_REMOVAL:
			Status	= SurpriseRemoval( DeviceObject, DeviceExtension, Irp );
			break;

		case IRP_MN_REMOVE_DEVICE:
			Status	= RemoveDevice( DeviceObject, DeviceExtension, Irp );
			break;

		default:
			//	IRP_MJ_PNP vʃhCoɑ
			Status	= SendRequest( DeviceObject, Irp );
			break;
	}

	//	QƃJE^Z
	DecrementReference( DeviceExtension );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	StartDevice
//----------------------------------------------------------------------------------------------
NTSTATUS StartDevice(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	ϐ錾
	NTSTATUS	Status;

	//	 IRP_MN_START_DEVICE vʃhCoɑ
	Status	= SendAndWaitRequest( DeviceObject, Irp );
	if( !NT_SUCCESS( Status ) )
	{
		CompleteRequest( Irp, Status, 0 );
		return( Status );
	}

	//	foCXLɂ
	Status	= EnableDevice( DeviceExtension );

	//	IRP_MN_START_DEVICE v
	CompleteRequest( Irp, Status, 0 );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	QueryCapabilities
//----------------------------------------------------------------------------------------------
NTSTATUS QueryCapabilities(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PIRP				Irp
	,IN PIO_STACK_LOCATION	CurrentIrpStack )
{
	//	ϐ錾
	NTSTATUS				Status;
	PDEVICE_CAPABILITIES	DeviceCapabilities;

	//	 IRP_MN_QUERY_CAPABILITIES vʃhCoɑ
	Status	= SendAndWaitRequest( DeviceObject, Irp );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	ʃhCoŐݒ肳ꂽfoCX̋@\ύX
	DeviceCapabilities	= CurrentIrpStack->Parameters.DeviceCapabilities.Capabilities;
	DeviceCapabilities->SurpriseRemovalOK	= TRUE;

	//	IRP_MN_QUERY_CAPABILITIES v
	CompleteRequest( Irp, Status, 0 );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	SurpriseRemoval
//----------------------------------------------------------------------------------------------
NTSTATUS SurpriseRemoval(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	foCX𖳌ɂ
	DisableDevice( DeviceExtension );

	//	IRP_MN_SURPRISE_REMOVAL vʃhCoɑ
	return( SendRequest( DeviceObject, Irp ) );
}

//----------------------------------------------------------------------------------------------
//	RemoveDevice
//----------------------------------------------------------------------------------------------
NTSTATUS RemoveDevice(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	foCX𖳌ɂ
	DisableDevice( DeviceExtension );

	//	IRP_MN_REMOVE_DEVICE vʃhCoɑ
	return( SendRequest( DeviceObject, Irp ) );
}

//----------------------------------------------------------------------------------------------
//	EnableDevice
//----------------------------------------------------------------------------------------------
NTSTATUS EnableDevice( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status;
	HANDLE		ThreadHandle;
	UCHAR		ThreadId;

	//	foCXLɂ
	if( DeviceExtension->DeviceState == Disable )
	{
		//	USB |[g
		Status	= SetupPort( DeviceExtension->NextDeviceObject );
		if( !NT_SUCCESS( Status ) )
		{
			return( Status );
		}

		//	foCXLqq擾
		Status	= GetDeviceDescriptor(
					 DeviceExtension->NextDeviceObject
					,&DeviceExtension->DeviceDescriptor );
		if( !NT_SUCCESS( Status ) )
		{
			return( Status );
		}

		//	RtBM[VLqq擾
		Status	= GetConfigurationDescriptor(
					 DeviceExtension->NextDeviceObject
					,&DeviceExtension->ConfigurationDescriptor );
		if( !NT_SUCCESS( Status ) )
		{
			return( Status );
		}

		//	C^[tFCXLqq擾
		Status	= GetInterfaceDescriptor(
					 DeviceExtension->ConfigurationDescriptor
					,&DeviceExtension->InterfaceDescriptor );
		if( !NT_SUCCESS( Status ) )
		{
			return( Status );
		}

		//	RtBM[VI
		Status	= SelectConfiguration(
					 DeviceExtension->NextDeviceObject
					,DeviceExtension->ConfigurationDescriptor
					,DeviceExtension->InterfaceDescriptor
					,DeviceExtension->InputPipeHandle
					,DeviceExtension->OutputPipeHandle );
		if( !NT_SUCCESS( Status ) )
		{
			return( Status );
		}

		//	ݒǂݍ܂
		for( ThreadId = 0; ThreadId < ControllerCount; ThreadId ++ )
		{
			DeviceExtension->ResetDevice[ThreadId]	= TRUE;
		}

		//	foCX̏ԂLɂ
		DeviceExtension->DeviceState	= Enable;
	}

	//	[J XbhJn
	if( DeviceExtension->ThreadStateIo == Disable )
	{
		DeviceExtension->ThreadStateIo	= Enable;
		for( ThreadId = 0; ThreadId < ControllerCount; ThreadId ++ )
		{
			//	Xbh p[^ݒ肷
			DeviceExtension->ThreadParam[ThreadId].ThreadId			= ThreadId;
			DeviceExtension->ThreadParam[ThreadId].DeviceExtension	= DeviceExtension;
			//	͗p[J XbhJn
			if( DeviceExtension->InputPipeHandle[ThreadId] != NULL )
			{
				//	͗p[J XbhJn
				PsCreateSystemThread(
					 &ThreadHandle
					,THREAD_ALL_ACCESS
					,NULL
					,NULL
					,NULL
					,WorkerThreadIn
					,&DeviceExtension->ThreadParam[ThreadId] );
				ObReferenceObjectByHandle(
					 ThreadHandle
					,THREAD_ALL_ACCESS
					,NULL
					,KernelMode
					,&DeviceExtension->ThreadObjectIn[ThreadId]
					,NULL );
				ZwClose( ThreadHandle );
			}
			//	o͗p[J XbhJn
			if( DeviceExtension->OutputPipeHandle[ThreadId] != NULL )
			{
				PsCreateSystemThread(
					 &ThreadHandle
					,THREAD_ALL_ACCESS
					,NULL
					,NULL
					,NULL
					,WorkerThreadOut
					,&DeviceExtension->ThreadParam[ThreadId] );
				ObReferenceObjectByHandle(
					 ThreadHandle
					,THREAD_ALL_ACCESS
					,NULL
					,KernelMode
					,&DeviceExtension->ThreadObjectOut[ThreadId]
					,NULL );
				ZwClose( ThreadHandle );
			}
		}
	}

	//	IRP p[J XbhJn
	if( DeviceExtension->ThreadStateIdc == Disable )
	{
		DeviceExtension->ThreadStateIdc	= Enable;
		PsCreateSystemThread(
			 &ThreadHandle
			,THREAD_ALL_ACCESS
			,NULL
			,NULL
			,NULL
			,WorkerThreadIdc
			,DeviceExtension );
		ObReferenceObjectByHandle(
			 ThreadHandle
			,THREAD_ALL_ACCESS
			,NULL
			,KernelMode
			,&DeviceExtension->ThreadObjectIdc
			,NULL );
		ZwClose( ThreadHandle );
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	WorkerThreadIn
//----------------------------------------------------------------------------------------------
VOID WorkerThreadIn( IN PVOID Context )
{
	//	ϐ錾
	NTSTATUS			Status;
	PTHREAD_PARAM		ThreadParam		= Context;
	PDEVICE_EXTENSION	DeviceExtension	= ThreadParam->DeviceExtension;
	UCHAR				ThreadId		= ThreadParam->ThreadId;
	SETTING				Setting;
	KIRQL				Irql;
	INPUT_REPORT		ControllerReport;
	LARGE_INTEGER		TimeOut;

	//	Xbh̗D揇ʂ
	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );

	for(;;)
	{
		//	Xbh̏I`FbN
		if( DeviceExtension->ThreadStateIo == Disable )
		{
			break;
		}

		//	Rg[[̐ݒǂݍ
		if(	DeviceExtension->ResetDevice[ThreadId]		== TRUE
		&&	DeviceExtension->OutputPipeHandle[ThreadId]	== NULL )
		{
			//	ݒǂݍ
			LoadSetting( &Setting );

			//	̓|[g̊荞݂֎~
			KeAcquireSpinLock( &DeviceExtension->LockReport, &Irql );

			//	ݒ𔽉f
			if( DeviceExtension->ResetDevice[ThreadId] == TRUE )
			{
				RtlCopyMemory( &DeviceExtension->Setting, &Setting, sizeof( SETTING ) );
				DeviceExtension->ResetDevice[ThreadId]	= FALSE;
			}

			//	̓|[g̊荞݂
			KeReleaseSpinLock( &DeviceExtension->LockReport, Irql );
		}

		//	Rg[[ |[g̓͂s
		Status	= InputReport(
					 DeviceExtension->NextDeviceObject
					,DeviceExtension->InputPipeHandle[ThreadId]
					,&ControllerReport
					,sizeof( INPUT_REPORT ) );
		if( NT_SUCCESS( Status ) )
		{
			//	Rg[[ڑꂽꍇ
			if(	ControllerReport.Type		== 0x08
			&&	ControllerReport.ParamType	== 0x80 )
			{
				DeviceExtension->Connected[ThreadId]			= TRUE;
				DeviceExtension->ConnectionReported[ThreadId]	= TRUE;
				DeviceExtension->ResetDevice[ThreadId]			= TRUE;
			}
			//	Rg[[ؒfꂽꍇ
			if(	ControllerReport.Type		== 0x08
			&&	ControllerReport.ParamType	== 0x00 )
			{
				DeviceExtension->Connected[ThreadId]			= FALSE;
				DeviceExtension->ConnectionReported[ThreadId]	= TRUE;
			}
			//	͗p|[g擾ꍇ
			if(	ControllerReport.ParamType				== 0x01
			&&	ControllerReport.Param1					== 0xF0
			&&	DeviceExtension->Connected[ThreadId]	== TRUE )
			{
				//	̓|[g̊荞݂֎~
				KeAcquireSpinLock( &DeviceExtension->LockReport, &Irql );

				//	Rg[[ |[g𔽉f
				RtlCopyMemory(
					 &DeviceExtension->ControllerReport[ThreadId]
					,&ControllerReport
					,sizeof( INPUT_REPORT ) );

				//	̓|[g̊荞݂
				KeReleaseSpinLock( &DeviceExtension->LockReport, Irql );

				//	|[gꂽޔ
				KeQuerySystemTime( &DeviceExtension->ControllerReportTime[ThreadId] );
			}
			TimeOut.QuadPart	= INTERVAL;
		} else {
			//	G[pCvZbg
			if( Status != STATUS_CANCELLED && Status != STATUS_DEVICE_NOT_CONNECTED )
			{
				ResetPipe(
					 DeviceExtension->NextDeviceObject
					,DeviceExtension->InputPipeHandle[ThreadId] );
			}
			TimeOut.QuadPart	= INTERVAL_STALL;
		}
		//	ҋ@
		KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
	}

	//	XbhJ
	ObDereferenceObject( DeviceExtension->ThreadObjectIn[ThreadId] );
	DeviceExtension->ThreadObjectIn[ThreadId]	= NULL;

	//	XbhIƂʒm
	KeSetEvent( &DeviceExtension->ThreadExitingIn[ThreadId], 0 ,FALSE );

	//	XbhI
	PsTerminateSystemThread( STATUS_SUCCESS );
}

//----------------------------------------------------------------------------------------------
//	WorkerThreadOut
//----------------------------------------------------------------------------------------------
VOID WorkerThreadOut( IN PVOID Context )
{
	//	ϐ錾
	NTSTATUS			Status;
	PTHREAD_PARAM		ThreadParam			= Context;
	PDEVICE_EXTENSION	DeviceExtension		= ThreadParam->DeviceExtension;
	UCHAR				ThreadId			= ThreadParam->ThreadId;
	SETTING				Setting;
	KIRQL				Irql;
	POUTPUT_REPORT		ConnectionReport	= &DeviceExtension->ConnectionReport[ThreadId];
	POUTPUT_REPORT		RingLightReport		= &DeviceExtension->RingLightReport[ThreadId];
	POUTPUT_REPORT		ActuatorsReport		= &DeviceExtension->ActuatorsReport[ThreadId];
	OUTPUT_REPORT		PrevActuatorsReport	= *ActuatorsReport;
	LONG				ReportCount			= 0;
	LARGE_INTEGER		TimeOut;

	//	Xbh̗D揇ʂ
	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );

	for(;;)
	{
		//	Xbh̏I`FbN
		if( DeviceExtension->ThreadStateIo == Disable )
		{
			break;
		}

		//	Rg[[̐ݒǂݍ
		if( DeviceExtension->ResetDevice[ThreadId] == TRUE )
		{
			//	ݒǂݍ
			LoadSetting( &Setting );

			//	̓|[g̊荞݂֎~
			KeAcquireSpinLock( &DeviceExtension->LockReport, &Irql );

			//	ݒ𔽉f
			if( DeviceExtension->ResetDevice[ThreadId] == TRUE )
			{
				RtlCopyMemory( &DeviceExtension->Setting, &Setting, sizeof( SETTING ) );
				DeviceExtension->ActuatorsReported[ThreadId]	= FALSE;
				DeviceExtension->RingLightReported[ThreadId]	= FALSE;
				DeviceExtension->ResetDevice[ThreadId]			= FALSE;
			}

			//	̓|[g̊荞݂
			KeReleaseSpinLock( &DeviceExtension->LockReport, Irql );
		}

		//	Rg[[̐ڑmF|[g̏o͂s
		if( DeviceExtension->ConnectionReported[ThreadId] == FALSE )
		{
			//	Rg[[̐ڑmF|[g쐬
			ConnectionReport->Type		= 0x08;
			ConnectionReport->ParamType	= 0x00;
			ConnectionReport->Param1	= 0x0F;
			ConnectionReport->Param2	= 0xC0;
			//	Rg[[̐ڑmF|[go͂
			Status	= OutputReport(
						 DeviceExtension->NextDeviceObject
						,DeviceExtension->OutputPipeHandle[ThreadId]
						,&DeviceExtension->ConnectionReport[ThreadId]
						,sizeof( OUTPUT_REPORT ) );
			if( NT_SUCCESS( Status ) )
			{
				;
			}
		}

		//	O Cg |[g̏o͂s
		if(	DeviceExtension->RingLightReported[ThreadId]	== FALSE
		&&	DeviceExtension->Connected[ThreadId]			== TRUE )
		{
			//	O Cg |[g쐬
			RingLightReport->Type		= 0x00;
			RingLightReport->ParamType	= 0x00;
			RingLightReport->Param1		= 0x08;
			if( DeviceExtension->Setting.RingLightPattern == 16 )
			{
				RingLightReport->Param2	= 0x42 + ThreadId;
			} else {
				RingLightReport->Param2	= 0x40
											+ (UCHAR)DeviceExtension->Setting.RingLightPattern;
			}
			//	O Cg |[go͂
			Status	= OutputReport(
						 DeviceExtension->NextDeviceObject
						,DeviceExtension->OutputPipeHandle[ThreadId]
						,&DeviceExtension->RingLightReport[ThreadId]
						,sizeof( OUTPUT_REPORT ) );
			if( NT_SUCCESS( Status ) )
			{
				DeviceExtension->RingLightReported[ThreadId]	= TRUE;
			}
		}

		//	AN`G[^ |[g쐬
		if( DeviceExtension->Connected[ThreadId] == TRUE )
		{
			ActuatorsReport->Type		= 0x00;
			ActuatorsReport->ParamType	= 0x01;
			ActuatorsReport->Param1		= 0x0F;
			ActuatorsReport->Param2		= 0xC0;
			if( DeviceExtension->ActuatorsReported[ThreadId] == FALSE )
			{
				ActuatorsReport->LeftActuator	= 0;
				ActuatorsReport->RightActuator	= 0;
			}

			//	̒lŃ|[g񐔂
			if(	ActuatorsReport->LeftActuator		!= PrevActuatorsReport.LeftActuator
			||	ActuatorsReport->RightActuator		!= PrevActuatorsReport.RightActuator
			||	DeviceExtension->ActuatorsReported[ThreadId]	== FALSE )
			{
				ReportCount	= 0;
			}

			//	AN`G[^ |[go͂
			if( ReportCount < 250 )
			{
				Status	= OutputReport(
							 DeviceExtension->NextDeviceObject
							,DeviceExtension->OutputPipeHandle[ThreadId]
							,&DeviceExtension->ActuatorsReport[ThreadId]
							,sizeof( OUTPUT_REPORT ) );
				if( NT_SUCCESS( Status ) )
				{
					//	|[g̓eޔ
					PrevActuatorsReport								= *ActuatorsReport;
					ReportCount ++;
					DeviceExtension->ActuatorsReported[ThreadId]	= TRUE;
					TimeOut.QuadPart								= INTERVAL;
				} else {
					//	G[pCvZbg
					if( Status != STATUS_CANCELLED && Status != STATUS_DEVICE_NOT_CONNECTED )
					{
						ResetPipe(
							 DeviceExtension->NextDeviceObject
							,DeviceExtension->OutputPipeHandle[ThreadId] );
					}
					TimeOut.QuadPart	= INTERVAL_STALL;
				}
			} else {
				TimeOut.QuadPart	= INTERVAL_WAIT;
			}
		} else {
			TimeOut.QuadPart	= INTERVAL_STALL;
		}
		//	ҋ@
		KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
	}

	//	XbhJ
	ObDereferenceObject( DeviceExtension->ThreadObjectOut[ThreadId] );
	DeviceExtension->ThreadObjectOut[ThreadId]	= NULL;

	//	XbhIƂʒm
	KeSetEvent( &DeviceExtension->ThreadExitingOut[ThreadId], 0 ,FALSE );

	//	XbhI
	PsTerminateSystemThread( STATUS_SUCCESS );
}

//----------------------------------------------------------------------------------------------
//	WorkerThreadIdc
//----------------------------------------------------------------------------------------------
VOID WorkerThreadIdc( IN PVOID Context )
{
	//	ϐ錾
	NTSTATUS			Status;
	PDEVICE_EXTENSION	DeviceExtension	= Context;
	KIRQL				Irql;
	PIRP				Irp;
	PIRP *				Irps;
	LARGE_INTEGER		TimeOut;

	//	Xbh̗D揇ʂ
	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );

	for(;;)
	{
		//	Xbh̏I`FbNiL[ IRP cĂꍇ͑Sďj
		if( DeviceExtension->ThreadStateIdc == Disable && DeviceExtension->IrpCount == 0 )
		{
			break;
		}

		//	rs
		KeAcquireSpinLock( &DeviceExtension->LockIrp, &Irql );

		//	L[ IRP 擾
		if( DeviceExtension->IrpCount > 0 )
		{
			//	L[ IRP 擾
			Irp	= DeviceExtension->Irps[0];
			//	擾 IRP L[菜
			if( DeviceExtension->IrpCount > 1 )
			{
				Irps	= (PIRP *)ExAllocatePoolWithTag(
							 NonPagedPool
							,(DeviceExtension->IrpCount - 1) * sizeof( PIRP )
							,'300T' );
				RtlCopyMemory(
					 Irps
					,&DeviceExtension->Irps[1]
					,(DeviceExtension->IrpCount - 1) * sizeof( PIRP ) );
			}
			if( DeviceExtension->Irps != NULL )
			{
				ExFreePool( DeviceExtension->Irps );
				DeviceExtension->Irps = NULL;
			}
			//	VL[ݒ肷
			DeviceExtension->Irps	= Irps;
			DeviceExtension->IrpCount --;
		} else {
			Irp	= NULL;
		}

		//	r
		KeReleaseSpinLock( &DeviceExtension->LockIrp, Irql );

		//	IRP 
		if( Irp != NULL )
		{
			ReadReport( Irp, DeviceExtension );
		} else {
			//	ҋ@
			TimeOut.QuadPart	= INTERVAL;
			KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
		}
	}

	//	XbhJ
	ObDereferenceObject( DeviceExtension->ThreadObjectIdc );
	DeviceExtension->ThreadObjectIdc	= NULL;

	//	XbhIƂʒm
	KeSetEvent( &DeviceExtension->ThreadExitingIdc, 0 ,FALSE );

	//	XbhI
	PsTerminateSystemThread( STATUS_SUCCESS );
}

//----------------------------------------------------------------------------------------------
//	DisableDevice
//----------------------------------------------------------------------------------------------
NTSTATUS DisableDevice( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	UCHAR	ThreadId;

	//	[J XbhÎ҂
	if( DeviceExtension->ThreadStateIo == Enable )
	{
		//	[J XbhI
		DeviceExtension->ThreadStateIo	= Disable;
		for( ThreadId = 0; ThreadId < ControllerCount; ThreadId ++ )
		{
			if( DeviceExtension->InputPipeHandle[ThreadId] != NULL )
			{
				//	͗ppCvA{[g
				AbortPipe(
					 DeviceExtension->NextDeviceObject
					,DeviceExtension->InputPipeHandle[ThreadId] );
				//	͗p[J Xbĥ҂
				KeWaitForSingleObject(
					 &DeviceExtension->ThreadExitingIn[ThreadId]
					,Executive
					,KernelMode
					,FALSE
					,NULL );
			}
			if( DeviceExtension->OutputPipeHandle[ThreadId] != NULL )
			{
				//	o͗ppCvA{[g
				AbortPipe(
					 DeviceExtension->NextDeviceObject
					,DeviceExtension->OutputPipeHandle[ThreadId] );
				//	o͗p[J Xbĥ҂
				KeWaitForSingleObject(
					 &DeviceExtension->ThreadExitingOut[ThreadId]
					,Executive
					,KernelMode
					,FALSE
					,NULL );
			}
		}
	}

	//	hCȍԂ𖳌ɂ
	DeviceExtension->DriverState	= Disable;

	//	IRP p[J Xbĥ҂
	if( DeviceExtension->ThreadStateIdc == Enable )
	{
		//	IRP p[J XbhI
		DeviceExtension->ThreadStateIdc	= Disable;

		//	IRP p[J Xbĥ҂
		KeWaitForSingleObject(
			 &DeviceExtension->ThreadExitingIdc
			,Executive
			,KernelMode
			,FALSE
			,NULL );
	}

	//	QƏICxg
	KeResetEvent( &DeviceExtension->ReferenceEnd );

	//	QƃJE^Z
	DecrementReference( DeviceExtension );
	DecrementReference( DeviceExtension );

	//	SĂ I/O vIAQƏIʒm܂őҋ@
	KeWaitForSingleObject( &DeviceExtension->ReferenceEnd, Executive, KernelMode, FALSE, NULL);

	//	foCX𖳌ɂ
	if( DeviceExtension->DeviceState == Enable )
	{
		//	RtBM[V̑I
		DeselectConfiguration( DeviceExtension->NextDeviceObject );

		//	RtBM[VLqqpobt@mۂĂΊJ
		if( DeviceExtension->ConfigurationDescriptor != NULL )
		{
			ExFreePool( DeviceExtension->ConfigurationDescriptor );
			DeviceExtension->ConfigurationDescriptor	= NULL;
		}

		//	C^[tFCXLqqpobt@mۂĂΊJ
		if( DeviceExtension->InterfaceDescriptor != NULL )
		{
			ExFreePool( DeviceExtension->InterfaceDescriptor );
			DeviceExtension->InterfaceDescriptor	= NULL;
		}

		//	foCX̏Ԃ𖳌ɂ
		DeviceExtension->DeviceState	= Disable;
	}

	//	hCȍԂLɂ
	DeviceExtension->DriverState	= Enable;

	//	QƃJE^̐킹
	IncrementReference( DeviceExtension );
	IncrementReference( DeviceExtension );

	return( STATUS_SUCCESS );
}
