
//	HID Minidriver for Microsoft Xbox 360 Wireless Controller

#include	"driver.h"

//----------------------------------------------------------------------------------------------
//	֐
//----------------------------------------------------------------------------------------------

NTSTATUS GetPortStatus( IN PDEVICE_OBJECT );
NTSTATUS ResetPort( IN PDEVICE_OBJECT );
NTSTATUS PipeRequest( IN PDEVICE_OBJECT, IN USHORT, IN USBD_PIPE_HANDLE );
NTSTATUS CallUSBDI( IN PDEVICE_OBJECT, IN ULONG, IN OUT PVOID );

//----------------------------------------------------------------------------------------------
//	SetupPort
//----------------------------------------------------------------------------------------------
NTSTATUS SetupPort( IN PDEVICE_OBJECT NextDeviceObject )
{
	//	ϐ錾
	NTSTATUS	Status;

	//	USB |[g̏Ԃ擾
	Status	= GetPortStatus( NextDeviceObject );
	if( !NT_SUCCESS( Status ) )
	{
		//	USB |[g
		Status	= ResetPort( NextDeviceObject );
		if( !NT_SUCCESS( Status ) )
		{
			return( Status );
		}

		//	ēx USB |[g̏Ԃ擾
		Status	= GetPortStatus( NextDeviceObject );
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	GetPortStatus
//----------------------------------------------------------------------------------------------
NTSTATUS GetPortStatus( IN PDEVICE_OBJECT NextDeviceObject )
{
	//	ϐ錾
	NTSTATUS	Status;
	ULONG		PortStatus;

	//	USB |[g̏Ԃ擾
	Status	= CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_GET_PORT_STATUS, &PortStatus );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	foCX USB |[gɐڑĂ邩ׂ
	if( ( PortStatus & USBD_PORT_CONNECTED ) != USBD_PORT_CONNECTED )
	{
		return( STATUS_NO_SUCH_DEVICE );
	}

	//	USB |[gLׂ
	if( ( PortStatus & USBD_PORT_ENABLED ) != USBD_PORT_ENABLED )
	{
		return( STATUS_DEVICE_NOT_READY );
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	ResetPort
//----------------------------------------------------------------------------------------------
NTSTATUS ResetPort( IN PDEVICE_OBJECT NextDeviceObject )
{
	//	USB |[gZbg
	return( CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_RESET_PORT, NULL ) );
}

//----------------------------------------------------------------------------------------------
//	GetDeviceDescriptor
//----------------------------------------------------------------------------------------------
NTSTATUS GetDeviceDescriptor(
	 IN PDEVICE_OBJECT			NextDeviceObject
	,OUT PUSB_DEVICE_DESCRIPTOR	DeviceDescriptor )
{
	//	ϐ錾
	NTSTATUS	Status;
	USHORT		UrbSize	= sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST );
	PURB		Urb;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePoolWithTag( NonPagedPool, UrbSize, '400T' );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	foCXLqq擾ׂ URB 쐬
	UsbBuildGetDescriptorRequest(
		 Urb
		,UrbSize
		,USB_DEVICE_DESCRIPTOR_TYPE
		,0
		,0
		,DeviceDescriptor
		,NULL
		,sizeof( USB_DEVICE_DESCRIPTOR )
		,NULL );

	//	foCXLqq擾ׂ URB 𔭍s
	Status	= CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		Status	= STATUS_UNSUCCESSFUL;
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	GetConfigurationDescriptor
//----------------------------------------------------------------------------------------------
NTSTATUS GetConfigurationDescriptor(
	 IN PDEVICE_OBJECT						NextDeviceObject
	,OUT PUSB_CONFIGURATION_DESCRIPTOR *	ConfigurationDescriptor )
{
	//	ϐ錾
	NTSTATUS	Status;
	USHORT		UrbSize	= sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST );
	PURB		Urb;
	LONG		ConfigurationDescriptorSize
						= sizeof( USB_CONFIGURATION_DESCRIPTOR );

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePoolWithTag( NonPagedPool, UrbSize, '500T' );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	RtBM[VLqq̓TCYmȂ̂ŒKv
	for(;;)
	{
		//	RtBM[VLqqpobt@쐬
		*ConfigurationDescriptor	= ExAllocatePoolWithTag(
										 NonPagedPool
										,ConfigurationDescriptorSize
										,'600T' );
		if( *ConfigurationDescriptor == NULL )
		{
			ExFreePool( Urb );
			return( STATUS_INSUFFICIENT_RESOURCES );
		}

		//	RtBM[VLqq擾ׂ URB 쐬
		UsbBuildGetDescriptorRequest(
			 Urb
			,UrbSize
			,USB_CONFIGURATION_DESCRIPTOR_TYPE
			,0
			,0
			,*ConfigurationDescriptor
			,NULL
			,ConfigurationDescriptorSize
			,NULL );

		//	RtBM[VLqq擾ׂ URB 𔭍s
		Status	= CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
		if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
		{
			Status	= STATUS_UNSUCCESSFUL;
			break;
		}

		//	쐬RtBM[VLqq̃TCYꍇAč쐬
		if( ConfigurationDescriptorSize < (*ConfigurationDescriptor)->wTotalLength )
		{
			ConfigurationDescriptorSize	= (*ConfigurationDescriptor)->wTotalLength;
			ExFreePool( *ConfigurationDescriptor );
			*ConfigurationDescriptor	= NULL;
		} else {
			break;
		}
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	GetInterfaceDescriptor
//----------------------------------------------------------------------------------------------
NTSTATUS GetInterfaceDescriptor(
	 IN PUSB_CONFIGURATION_DESCRIPTOR	ConfigurationDescriptor
	,OUT PUSB_INTERFACE_DESCRIPTOR * *	InterfaceDescriptor )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;
	LONG		Index;

	//	C^[tFCXLqqpobt@쐬
	*InterfaceDescriptor	= ExAllocatePoolWithTag(
								 NonPagedPool
								,ConfigurationDescriptor->bNumInterfaces
									* sizeof( PUSB_INTERFACE_DESCRIPTOR )
								,'700T' );
	if( *InterfaceDescriptor == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	SẴC^[tFCXLqq擾
	(*InterfaceDescriptor)[0]	= (PUSB_INTERFACE_DESCRIPTOR)ConfigurationDescriptor;
	for( Index = 0; Index < ConfigurationDescriptor->bNumInterfaces; Index ++ )
	{
		//	C^[tFCXLqq擾
		(*InterfaceDescriptor)[Index]	= USBD_ParseConfigurationDescriptorEx(
											 ConfigurationDescriptor
											,(*InterfaceDescriptor)[max( 0, Index - 1 )]
											,Index
											,0
											,-1
											,-1
											,-1 );
		if( (*InterfaceDescriptor)[Index] == NULL )
		{
			Status	= STATUS_UNSUCCESSFUL;
			break;
		}
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	SelectConfiguration
//----------------------------------------------------------------------------------------------
NTSTATUS SelectConfiguration(
	 IN PDEVICE_OBJECT					NextDeviceObject
	,IN PUSB_CONFIGURATION_DESCRIPTOR	ConfigurationDescriptor
	,IN PUSB_INTERFACE_DESCRIPTOR *		InterfaceDescriptor
	,OUT USBD_PIPE_HANDLE *				InputPipeHandle
	,OUT USBD_PIPE_HANDLE *				OutputPipeHandle )
{
	//	ϐ錾
	NTSTATUS					Status;
	LONG						InterfaceListEntrySize
									= ( ConfigurationDescriptor->bNumInterfaces + 1 )
										* sizeof( USBD_INTERFACE_LIST_ENTRY );
	PUSBD_INTERFACE_LIST_ENTRY	InterfaceListEntry;
	LONG						Index;
	PURB						Urb;
	PUSBD_INTERFACE_INFORMATION	InterfaceInformation;

	//	C^[tFCX Xgpobt@쐬
	InterfaceListEntry	= ExAllocatePoolWithTag(
							  NonPagedPool
							 ,InterfaceListEntrySize
							 ,'800T' );
	if( InterfaceListEntry == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	C^[tFCX Xgݒ肷
	RtlZeroMemory( InterfaceListEntry, InterfaceListEntrySize );
	for( Index = 0; Index < ConfigurationDescriptor->bNumInterfaces; Index ++ )
	{
		InterfaceListEntry[Index].InterfaceDescriptor	= InterfaceDescriptor[Index];
	}

	//	RtBM[VI邽߂ URB 쐬
	Urb	= USBD_CreateConfigurationRequestEx(
			 ConfigurationDescriptor
			,InterfaceListEntry );
	if( Urb == NULL )
	{
		ExFreePool( InterfaceListEntry );
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	RtBM[VIׂ URB 𔭍s
	Status	= CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( NT_SUCCESS( Status ) )
	{
		for( Index = 0; Index < ConfigurationDescriptor->bNumInterfaces; Index += 2 )
		{
			//	Rg[[葽ꍇ͖
			if( ControllerCount < Index / 2)
			{
				break;
			}

			//	C^[tFCX CtH[V擾
			InterfaceInformation	= InterfaceListEntry[Index].Interface;
			//	pCv擾
			switch( InterfaceInformation->NumberOfPipes )
			{
				case 0:
					Status	= STATUS_UNSUCCESSFUL;
					break;

				case 1:
					InputPipeHandle[Index / 2]	= InterfaceInformation->Pipes[0].PipeHandle;
					break;

				default:
					InputPipeHandle[Index / 2]	= InterfaceInformation->Pipes[0].PipeHandle;
					OutputPipeHandle[Index / 2]	= InterfaceInformation->Pipes[1].PipeHandle;
					break;
			}
		}
	}

	//	C^[tFCX Xgpobt@
	ExFreePool( InterfaceListEntry );

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	DeselectConfiguration
//----------------------------------------------------------------------------------------------
NTSTATUS DeselectConfiguration( IN PDEVICE_OBJECT NextDeviceObject )
{
	//	ϐ錾
	NTSTATUS	Status;
	USHORT		UrbSize	= sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST );
	PURB		Urb;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePoolWithTag( NonPagedPool, UrbSize, '900T' );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	RtBM[VIׂ URB 쐬
	UsbBuildSelectConfigurationRequest( Urb, UrbSize, NULL );

	//	RtBM[VIׂ URB 𔭍s
	Status	= CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		ExFreePool( Urb );
		return( STATUS_UNSUCCESSFUL );
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	ResetPipe
//----------------------------------------------------------------------------------------------
NTSTATUS ResetPipe(
	 IN PDEVICE_OBJECT		NextDeviceObject
	,IN USBD_PIPE_HANDLE	PipeHandle )
{
	//	pCvZbg
	return( PipeRequest( NextDeviceObject, URB_FUNCTION_RESET_PIPE, PipeHandle ) );
}

//----------------------------------------------------------------------------------------------
//	AbortPipe
//----------------------------------------------------------------------------------------------
NTSTATUS AbortPipe(
	 IN PDEVICE_OBJECT		NextDeviceObject
	,IN USBD_PIPE_HANDLE	PipeHandle )
{
	//	pCvA{[g
	return( PipeRequest( NextDeviceObject, URB_FUNCTION_ABORT_PIPE, PipeHandle ) );
}

//----------------------------------------------------------------------------------------------
//	PipeRequest
//----------------------------------------------------------------------------------------------
NTSTATUS PipeRequest(
	 IN PDEVICE_OBJECT		NextDeviceObject
	,IN USHORT				Function
	,IN USBD_PIPE_HANDLE	PipeHandle )
{
	//	ϐ錾
	NTSTATUS	Status;
	USHORT		UrbSize	= sizeof( struct _URB_PIPE_REQUEST );
	PURB		Urb;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePoolWithTag( NonPagedPool, UrbSize, '010T' );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	pCvւ URB 쐬
	Urb->UrbHeader.Length			= (USHORT)sizeof( struct _URB_PIPE_REQUEST );
	Urb->UrbHeader.Function			= Function;
	Urb->UrbPipeRequest.PipeHandle	= PipeHandle;

	//	pCvւ URB 𔭍s
	Status	= CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		Status	= STATUS_UNSUCCESSFUL;
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	InputReport
//----------------------------------------------------------------------------------------------
NTSTATUS InputReport(
	 IN PDEVICE_OBJECT		NextDeviceObject
	,IN USBD_PIPE_HANDLE	PipeHandle
	,OUT INPUT_REPORT *		InputReport
	,IN LONG				InputReportSize )
{
	//	ϐ錾
	NTSTATUS	Status;
	USHORT		UrbSize	= sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER );
	PURB		Urb;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePoolWithTag( NonPagedPool, UrbSize, '110T' );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	̓|[g擾ׂ URB 쐬
	UsbBuildInterruptOrBulkTransferRequest(
		 Urb
		,UrbSize
		,PipeHandle
		,InputReport
		,NULL
		,InputReportSize
		,USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK
		,NULL );

	//	̓|[g擾ׂ URB 𔭍s
	Status	= CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		Status	= STATUS_UNSUCCESSFUL;
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	OutputReport
//----------------------------------------------------------------------------------------------
NTSTATUS OutputReport(
	 IN PDEVICE_OBJECT		NextDeviceObject
	,IN USBD_PIPE_HANDLE	PipeHandle
	,OUT OUTPUT_REPORT *	OutputReport
	,IN LONG				OutputReportSize )
{
	//	ϐ錾
	NTSTATUS	Status;
	USHORT		UrbSize	= sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER );
	PURB		Urb;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePoolWithTag( NonPagedPool, UrbSize, '210T' );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	o̓|[gݒ肷ׂ URB 쐬
	UsbBuildInterruptOrBulkTransferRequest(
		 Urb
		,UrbSize
		,PipeHandle
		,OutputReport
		,NULL
		,OutputReportSize
		,USBD_SHORT_TRANSFER_OK
		,NULL );
	//	o̓|[gݒ肷ׂ URB 𔭍s
	Status	= CallUSBDI( NextDeviceObject, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		Status	= STATUS_UNSUCCESSFUL;
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	CallUSBDI
//----------------------------------------------------------------------------------------------
NTSTATUS CallUSBDI(
	 IN PDEVICE_OBJECT	NextDeviceObject
	,IN ULONG			IOCTLCode
	,IN OUT PVOID		Arg1 )
{
	//	ϐ錾
	NTSTATUS			Status;
	PIRP				Irp;
	KEVENT				WaitEvent;
	IO_STATUS_BLOCK		IoStatus;
	PIO_STACK_LOCATION	NextIrpStack;

	//	I/O v쐬
	KeInitializeEvent( &WaitEvent, NotificationEvent, FALSE );
	Irp	= IoBuildDeviceIoControlRequest(
			 IOCTLCode
			,NextDeviceObject
			,NULL
			,0
			,NULL
			,0
			,TRUE
			,&WaitEvent
			,&IoStatus );
	NextIrpStack	= IoGetNextIrpStackLocation( Irp );
	NextIrpStack->Parameters.Others.Argument1	= Arg1;

	//	ʃhCo I/O v𔭍s
	Status	= IoCallDriver( NextDeviceObject, Irp );
	if( Status == STATUS_PENDING )
	{
		//	ʃhCȍ܂őҋ@
		KeWaitForSingleObject(
			 &WaitEvent
			,Executive
			,KernelMode
			,FALSE
			,NULL );
		Status	= IoStatus.Status;
	}

	return( Status );
}
