
//	Property Sheet for Microsoft Xbox 360 Wireless Controller

#include	"device.h"

//----------------------------------------------------------------------------------------------
//	CDevice
//----------------------------------------------------------------------------------------------
CDevice::CDevice()
{
	//	T|[gĂfoCX̐擾
	SupportedDeviceCount	= 0;
	GetRegistry(
		 &SupportedDeviceCount
		,REG_DWORD
		,sizeof( LONG )
		,SUPPORTED_DEVICE_COUNT );
	SupportedDeviceCount	= max( 0, min( SupportedDeviceCount, 256 ) );

	//	T|[gĂfoCX̑擾
	if( SupportedDeviceCount > 0 )
	{
		SupportedDeviceAttribute	= (ULONG *)LocalAlloc(
										  LPTR
										 ,SupportedDeviceCount * sizeof( ULONG ) );
		for( LONG Index = 0; Index < SupportedDeviceCount; Index ++ )
		{
			//	WXgfoCX̑擾
			WCHAR	Attribute[32];
			swprintf_s( Attribute, SUPPORTED_DEVICE_ATTRIBUTE, Index );
			GetRegistry(
				 &SupportedDeviceAttribute[Index]
				,REG_DWORD
				,sizeof( ULONG )
				,Attribute );
		}
	} else {
		SupportedDeviceAttribute	= NULL;
	}
}

//----------------------------------------------------------------------------------------------
//	~CDevice
//----------------------------------------------------------------------------------------------
CDevice::~CDevice()
{
	//	T|[gĂfoCX̑J
	if( SupportedDeviceAttribute != NULL )
	{
		LocalFree( SupportedDeviceAttribute );
	}
}

//----------------------------------------------------------------------------------------------
//	GetRegistry
//----------------------------------------------------------------------------------------------
VOID CDevice::GetRegistry(
	 VOID *		Value
	,ULONG		ValueType
	,ULONG		ValueSize
	,WCHAR *	ValueName )
{
	//	ϐ錾
	HKEY	Key;
	DWORD	Disposition;
	LONG	Result;

	//	WXgJ
	Result	= RegCreateKeyEx(
				 HKEY_LOCAL_MACHINE
				,REGISTRY_KEY
				,0
				,NULL
				,REG_OPTION_NON_VOLATILE
				,KEY_ALL_ACCESS
				,NULL
				,&Key
				,&Disposition  );
	if( Result == ERROR_SUCCESS )
	{
		//	WXgǂݍ
		RegQueryValueEx( Key, ValueName, NULL, &ValueType, (LPBYTE)Value, &ValueSize );
		//	WXg
		RegCloseKey( Key );
	}
}

//----------------------------------------------------------------------------------------------
//	IsConnected
//----------------------------------------------------------------------------------------------
BOOL CDevice::IsConnected()
{
	//	foCXڑĂ邩ǂԂ
	return( IoDevice( IO_MODE_IS_CONNECTED, NULL, NULL ) );
}

//----------------------------------------------------------------------------------------------
//	GetState
//----------------------------------------------------------------------------------------------
BOOL CDevice::GetState( FEATURE_REPORT_ID * FeatureReport )
{
	//	foCX̏Ԃ擾
	return( IoDevice( IO_MODE_GET_STATE, NULL, FeatureReport ) );
}

//----------------------------------------------------------------------------------------------
//	SetState
//----------------------------------------------------------------------------------------------
BOOL CDevice::SetState( HID_WRITE_REPORT_ID * WriteReport )
{
	//	foCX̏Ԃݒ肷
	return( IoDevice( IO_MODE_SET_STATE, WriteReport, NULL ) );
}

//----------------------------------------------------------------------------------------------
//	Apply
//----------------------------------------------------------------------------------------------
BOOL CDevice::Apply()
{
	//	foCXɕύXʒm
	return( IoDevice( IO_MODE_APPLY, NULL, NULL ) );
}

//----------------------------------------------------------------------------------------------
//	IoDevice
//----------------------------------------------------------------------------------------------
BOOL CDevice::IoDevice(
	 LONG					IoMode
	,HID_WRITE_REPORT_ID *	WriteReport
	,FEATURE_REPORT_ID *	FeatureReport )
{
	//	ϐ錾
	GUID				GUID_HID;
	LONG				InterfaceIndex	= 0;
	BOOL				DeviceFound		= FALSE;
	FEATURE_REPORT_ID	Report;

	//	foCX NX̃nh擾
	HidD_GetHidGuid( &GUID_HID );
	HDEVINFO	DevInfo	= SetupDiGetClassDevs(
							 &GUID_HID
							,NULL
							,NULL
							,DIGCF_PRESENT | DIGCF_INTERFACEDEVICE );
	if( DevInfo == INVALID_HANDLE_VALUE )
	{
		return( FALSE );
	}

	//	Ώۂ̃C^[tFCXT
	for(;;)
	{
		//	C^[tFCX擾
		SP_INTERFACE_DEVICE_DATA	Interface;
		Interface.cbSize	= sizeof( SP_INTERFACE_DEVICE_DATA );
		BOOL	Result	= SetupDiEnumDeviceInterfaces(
							 DevInfo
							,NULL
							,&GUID_HID
							,InterfaceIndex
							,&Interface );
		if( Result == FALSE )
		{
			SetupDiDestroyDeviceInfoList( DevInfo );
			return( DeviceFound );
		}

		//	C^[tFCX̏ڍ׏̃TCY擾
		DWORD	DetailSize;
		SetupDiGetDeviceInterfaceDetail( DevInfo, &Interface, NULL, 0, &DetailSize, NULL );

		//	C^[tFCX̏ڍ׏擾
		PSP_INTERFACE_DEVICE_DETAIL_DATA	Detail;
		Detail	= (PSP_INTERFACE_DEVICE_DETAIL_DATA)LocalAlloc( LPTR, DetailSize );
		if( Detail == NULL )
		{
			SetupDiDestroyDeviceInfoList( DevInfo );
			return( DeviceFound );
		}
		Detail->cbSize	= sizeof( SP_INTERFACE_DEVICE_DETAIL_DATA );
		Result	= SetupDiGetDeviceInterfaceDetail(
					 DevInfo
					,&Interface
					,Detail
					,DetailSize
					,NULL
					,NULL );
		if( Result == FALSE )
		{
			SetupDiDestroyDeviceInfoList( DevInfo );
			LocalFree( Detail );
			return( DeviceFound );
		}

		//	foCX C^[tFCXJ
		HANDLE	Device	= CreateFile(
							 Detail->DevicePath
							,GENERIC_READ | GENERIC_WRITE
							,FILE_SHARE_READ | FILE_SHARE_WRITE
							,NULL
							,OPEN_EXISTING
							,FILE_ATTRIBUTE_NORMAL
							,NULL );
		if( Device == INVALID_HANDLE_VALUE )
		{
			LocalFree( Detail );
			InterfaceIndex	++;
			continue;
		}

		//	C^[tFCX̏ڍׂɃx_ IDAv_Ng ID ܂܂邩`FbN
		HIDD_ATTRIBUTES	HidAttribute;
		HidD_GetAttributes( Device, &HidAttribute );
		BOOL	TargetDeviceFound	= FALSE;
		for( LONG DeviceIndex = 0; DeviceIndex < SupportedDeviceCount; DeviceIndex ++ )
		{
			if(	HidAttribute.VendorID	== HIWORD( SupportedDeviceAttribute[DeviceIndex] )
			&&	HidAttribute.ProductID	== LOWORD( SupportedDeviceAttribute[DeviceIndex] ) )
			{
				TargetDeviceFound	= TRUE;
				break;
			}
		}

		//	foCX𔭌ꍇ
		if( TargetDeviceFound == TRUE )
		{
			switch( IoMode )
			{
				case IO_MODE_IS_CONNECTED:
					;
					break;

				case IO_MODE_GET_STATE:
					//	foCX̏Ԃ擾
					ZeroMemory( &Report, sizeof( FEATURE_REPORT_ID ) );
					Report.Id	= GetControllerId( Detail->DevicePath );
					Result	= HidD_GetFeature( Device, &Report, sizeof( FEATURE_REPORT_ID ) );
					if( Result == TRUE )
					{
						//	Rg[ɑΏ邽߁Af[^Z
						//	fW^ {^
						FeatureReport->DigitalButton	|= Report.DigitalButton;
						//	gK[
						FeatureReport->LeftTrigger		= ADD_UCHAR(
															 FeatureReport->LeftTrigger
															,Report.LeftTrigger );
						//	EgK[
						FeatureReport->RightTrigger		= ADD_UCHAR(
															 FeatureReport->RightTrigger
															,Report.RightTrigger );
						//	XeBbN X
						FeatureReport->LeftStickX		= ADD_SHORT(
															 FeatureReport->LeftStickX
															,Report.LeftStickX );
						//	XeBbN Y
						FeatureReport->LeftStickY		= ADD_SHORT(
															 FeatureReport->LeftStickY
															,Report.LeftStickY );
						//	XeBbN X
						FeatureReport->RightStickX		= ADD_SHORT(
															 FeatureReport->RightStickX
															,Report.RightStickX );
						//	XeBbN Y
						FeatureReport->RightStickY		= ADD_SHORT(
															 FeatureReport->RightStickY
															,Report.RightStickY );
					}
					break;

				case IO_MODE_SET_STATE:
					//	foCX̏Ԃݒ肷
					DWORD	ReqLen;
					WriteReport->Id	= GetControllerId( Detail->DevicePath );
					WriteFile(
						 Device
						,WriteReport
						,sizeof( HID_WRITE_REPORT_ID )
						,&ReqLen
						,NULL );
					break;

				case IO_MODE_APPLY:
					//	foCXɕύXʒm
					ZeroMemory( &Report, sizeof( FEATURE_REPORT_ID ) );
					Report.Id	= GetControllerId( Detail->DevicePath );
					HidD_SetFeature( Device, &Report, sizeof( FEATURE_REPORT_ID ) );
					break;
			}
			DeviceFound	= TRUE;
		}

		//	foCX C^[tFCX
		CloseHandle( Device );

		//	̃C^[tFCX
		LocalFree( Detail );
		InterfaceIndex	++;
	}

	//	foCX NX̃nhJ
	SetupDiDestroyDeviceInfoList( DevInfo );

	return( DeviceFound );
}

//----------------------------------------------------------------------------------------------
//	GetControllerId
//----------------------------------------------------------------------------------------------
CHAR CDevice::GetControllerId( WCHAR * DevicePath )
{
	//	ϐ錾
	CHAR	ControllerId;
	WCHAR *	ControllerIdPos;

	//	Rg[ ID 擾
	ControllerIdPos	= wcsstr( DevicePath, L"&col0" );
	if( ControllerIdPos == NULL )
	{
		ControllerId	= 1;
	} else {
		ControllerId	= ControllerIdPos[5] - '0';
	}

	return ControllerId;
}
